@@ -206,11 +206,30 @@
     Updated by      Richard Gooch   18-AUG-2001: Dynamically load libnsl at
   run-time as needed, rather than linking. Based on patch from Adam J. Richter.
 
-    Last updated by Richard Gooch   19-AUG-2001: Switched to stat(2) in
+    Updated by      Richard Gooch   19-AUG-2001: Switched to stat(2) in
   <action_permissions> to avoid harmless false positives. Return 0 instead of
   exiting if passwd or group entry not found in database. Do not exit in
   <action_execute> if fork(2) fails.
 
+    Updated by      Richard Gooch   16-SEP-2001: Set environment to "safe mode"
+  when invoking modprobe. Generate error message if /lib/modutils.so doesn't
+  contain "modprobe" symbol. Add debugging message if /lib/modutils.so missing.
+  Based on patch from Russell Coker.
+
+    Updated by      Richard Gooch   6-NOV-2001: Called setsid(2) when closing
+  fds[0:2] so that hangups from the old controlling terminal are ignored.
+
+    Updated by      Richard Gooch   22-NOV-2001: Moved call to setsid(2) to
+  immediately after forking child/daemon. Added call to setpgid(2) if running
+  in the foreground. These changes are in anticipation of a kernel switch from
+  ancestor process checking to process group checking.
+
+    Updated by      Richard Gooch   25-NOV-2001: Created <<event_types>> table.
+  Generate synthetic REGISTER event for directories in <do_scan_and_service>.
+
+    Last updated by Richard Gooch   28-NOV-2001: Renamed <make_symlink> to
+  <mksymlink> and made it global. Changed to return 0 on "success".
+
 
 */
 #define _GNU_SOURCE
@@ -243,21 +262,17 @@
 #include "devfsd.h"
 #include "version.h"
 
-#ifndef RTLD_DEFAULT  /*  Glibc 2.2 broke old code  */
-#  define RTLD_DEFAULT RTLD_NEXT
+#ifndef RTLD_DEFAULT  /*  Libc 5 doesn't define it, but it works  */
+#  define RTLD_DEFAULT NULL
 #endif
 
-#define CONFIG_FILE "/etc/devfsd.conf"
-#define MAX_ARGS (6 + 1)
+#define CONFIG_FILE  "/etc/devfsd.conf"
+#define MAX_ARGS     (6 + 1)
 #define MAX_SUBEXPR  10
 
 /*  Update only after changing code to reflect new protocol  */
 #define DEVFSD_PROTOCOL_REVISION_DAEMON  5
 
-#ifndef DEVFSD_NOTIFY_DELETE  /*  Temporary hack  */
-#  define DEVFSD_NOTIFY_DELETE        7
-#endif
-
 /*  Compile-time check  */
 #if DEVFSD_PROTOCOL_REVISION_KERNEL != DEVFSD_PROTOCOL_REVISION_DAEMON
 #error protocol version mismatch. Update your kernel headers
@@ -359,6 +374,10 @@ EXTERN_FUNCTION (const char *get_old_name,
 flag syslog_is_open = FALSE;
 
 
+/*  Public functions  */
+int mksymlink (const char *oldpath, const char *newpath);
+
+
 /*  Private functions  */
 static void setup_initial_entries ();
 static void read_config (CONST char *location, flag optional,
@@ -393,11 +412,10 @@ static void do_debug (int fd);
 static uid_t get_uid (const char *string);
 static gid_t get_gid (const char *string);
 static mode_t get_mode (const char *string);
-static void sig_handler (int sig);
+static void sighup_handler (int sig);
 static CONST char *get_variable (CONST char *variable, void *info);
 static void do_open_syslog ();
 static void do_scan_and_service (CONST char *dirname);
-static flag make_symlink (const char *oldpath, const char *newpath);
 static flag make_dir_tree (const char *path);
 static struct shared_object *get_shared_object (CONST char *name);
 static flag expand_expression(char *output, unsigned int outsize,
@@ -437,6 +455,23 @@ static struct initial_symlink_struct
 static int (*my_yp_all) (char *domain, char *map,
 			 struct ypall_callback *callback); /* = NULL */
 static int (*my_yp_get_default_domain) (char **domainptr); /* = NULL */
+static struct event_type
+{
+    unsigned int type;        /*  The DEVFSD_NOTIFY_* value                  */
+    CONST char *config_name;  /*  The name used in the config file           */
+    CONST char *debug_name;   /*  The name used when debugging the protocol  */
+} event_types[] =
+{
+    {DEVFSD_NOTIFY_REGISTERED,   "REGISTER",   "registered"},
+    {DEVFSD_NOTIFY_UNREGISTERED, "UNREGISTER", "unregistered"},
+    {DEVFSD_NOTIFY_ASYNC_OPEN,   "ASYNC_OPEN", "asynchronously opened"},
+    {DEVFSD_NOTIFY_CLOSE,        "CLOSE",      "closed"},
+    {DEVFSD_NOTIFY_LOOKUP,       "LOOKUP",     "lookup"},
+    {DEVFSD_NOTIFY_CHANGE,       "CHANGE",     "changed"},
+    {DEVFSD_NOTIFY_CREATE,       "CREATE",     "created"},
+    {DEVFSD_NOTIFY_DELETE,       "DELETE",     "deleted"},
+    {0xffffffff,                 NULL,         NULL}
+};
 
 
 /*  Public functions follow  */
@@ -527,7 +562,7 @@ int main (int argc, char **argv)
 	fprintf (stderr, "devfsd: Linux device filesystem daemon v%s\n",
 		 DEVFSD_VERSION);
 	fprintf (stderr,
-		 "(C) 1998-2000  Richard Gooch <rgooch@atnf.csiro.au>\n\n");
+		 "(C) 1998-2001  Richard Gooch <rgooch@atnf.csiro.au>\n\n");
 	fprintf (stderr, "Daemon protocol revision:\t%d\n",
 		 DEVFSD_PROTOCOL_REVISION_DAEMON);
 	fprintf (stderr, "Kernel-side protocol revision:\t%d\n", proto_rev);
@@ -557,7 +592,7 @@ int main (int argc, char **argv)
     sigemptyset (&new_action.sa_mask);
     new_action.sa_flags = 0;
     /*  Set up SIGHUP handler  */
-    new_action.sa_handler = sig_handler;
+    new_action.sa_handler = sighup_handler;
     if (sigaction (SIGHUP, &new_action, NULL) != 0)
     {
 	fprintf (stderr, "Error setting SIGHUP handler\t%s\n", ERRSTRING);
@@ -598,7 +633,9 @@ int main (int argc, char **argv)
 	    exit (0);
 	    /*break;*/
 	}
+	setsid ();        /*  Prevent hangups and become pgrp leader         */
     }
+    else setpgid (0, 0);  /*  Become process group leader                    */
     do_open_syslog ();
     while (TRUE)
     {
@@ -772,23 +809,13 @@ static void process_config_line (CONST char *line, unsigned long *event_mask)
 	exit (1);
     }
     memset (new, 0, sizeof *new);
-    if (strcasecmp (when, "REGISTER") == 0)
-	new->action.when = DEVFSD_NOTIFY_REGISTERED;
-    else if (strcasecmp (when, "UNREGISTER") == 0)
-	new->action.when = DEVFSD_NOTIFY_UNREGISTERED;
-    else if (strcasecmp (when, "ASYNC_OPEN") == 0)
-	new->action.when = DEVFSD_NOTIFY_ASYNC_OPEN;
-    else if (strcasecmp (when, "CLOSE") == 0)
-	new->action.when = DEVFSD_NOTIFY_CLOSE;
-    else if (strcasecmp (when, "LOOKUP") == 0)
-	new->action.when = DEVFSD_NOTIFY_LOOKUP;
-    else if (strcasecmp (when, "CHANGE") == 0)
-	new->action.when = DEVFSD_NOTIFY_CHANGE;
-    else if (strcasecmp (when, "CREATE") == 0)
-	new->action.when = DEVFSD_NOTIFY_CREATE;
-    else if (strcasecmp (when, "DELETE") == 0)
-	new->action.when = DEVFSD_NOTIFY_DELETE;
-    else
+    for (count = 0; event_types[count].config_name != NULL; ++count)
+    {
+	if (strcasecmp (when, event_types[count].config_name) != 0) continue;
+	new->action.when = event_types[count].type;
+	break;
+    }
+    if (event_types[count].config_name == NULL)
     {
 	SYSLOG (LOG_ERR, "bad WHEN in config line: \"%s\"\n", line);
 	SYSLOG (LOG_ERR, "exiting\n");
@@ -1084,7 +1111,7 @@ static void action_permissions (const struct devfsd_notify_struct *info,
 
     if (stat (info->devname, &statbuf) != 0)
     {
-	SYSLOG (LOG_ERR, "error lstat(2)ing: \"%s\"\t%s\n",
+	SYSLOG (LOG_ERR, "error stat(2)ing: \"%s\"\t%s\n",
 		info->devname, ERRSTRING);
 	return;
     }
@@ -1133,6 +1160,8 @@ static void action_modload (const struct devfsd_notify_struct *info,
 {
     char *argv[6];
     char device[STRING_LENGTH];
+    char *env[4] = {"HOME=/", "TERM=linux",  /*  Safe mode environment  */
+		    "PATH=/sbin:/usr/sbin:/bin:/usr/bin", NULL};
     static int (*modprobe) (int argc, char **argv) = NULL;
     static flag first_time = TRUE;
 
@@ -1142,7 +1171,12 @@ static void action_modload (const struct devfsd_notify_struct *info,
 
 	first_time = FALSE;
 	if ( ( lib = dlopen ("/lib/modutils.so", RTLD_NOW) ) != NULL )
-	    modprobe = dlsym (lib, "modprobe");
+	{
+	    if ( ( modprobe = dlsym (lib, "modprobe") ) == NULL )
+		SYSLOG(LOG_ERR, "/lib/modutils.so doesn't contain modprobe()");
+	}
+	else if (trace_level > 1)
+	    fprintf (stderr, "/lib/modutils.so load failure: %s\n",dlerror() );
     }
     argv[0] = "/sbin/modprobe";
     argv[1] = "-k";
@@ -1152,10 +1186,15 @@ static void action_modload (const struct devfsd_notify_struct *info,
     argv[5] = NULL;
     snprintf (device, sizeof (device), "/dev/%s", info->devname);
     if (trace_level > 1)
-	fprintf (stderr, "Calling modprobe with name: \"%s\"\n", device);
+	fprintf (stderr, "%s modprobe with name: \"%s\"\n",
+		 modprobe ? "Calling" : "Executing", device);
     if (modprobe != NULL)
     {
+	char **oldenv = environ;
+
+	environ = env;
 	(*modprobe) (5, argv);
+	environ = oldenv;
 	return;
     }
     switch ( fork () )
@@ -1174,6 +1213,7 @@ static void action_modload (const struct devfsd_notify_struct *info,
 	return;
 	/*break;*/
     }
+    environ = env;
     execvp (argv[0], argv);
     SYSLOG (LOG_ERR, "error execing: \"%s\"\t%s\n", argv[0], ERRSTRING);
     _exit (1);
@@ -1518,7 +1558,7 @@ static void action_compat (const struct devfsd_notify_struct *info,
     {
       case AC_MKOLDCOMPAT:
       case AC_MKNEWCOMPAT:
-	if ( make_symlink (dest_name, compat_name) )
+	if (mksymlink (dest_name, compat_name) == 0)
 	    if (trace_level > 1)
 		fprintf (stderr, "made symlink: \"%s\" for dev: %u,%u\n",
 			compat_name, info->major, info->minor);
@@ -1590,7 +1630,6 @@ static void do_debug (int fd)
 {
     ssize_t bytes;
     struct devfsd_notify_struct info;
-    char *type;
 
     fprintf (stderr, "Running in debug mode: no actions will be taken...\n");
     /*  Tell devfs what events we care about  */
@@ -1601,6 +1640,9 @@ static void do_debug (int fd)
     }
     while ( ( bytes = read (fd, (char *) &info, sizeof info) ) >= 0 )
     {
+	unsigned int count;
+	CONST char *type = NULL;
+
 	if (bytes < 1)
 	{
 	    fprintf (stderr, "devfs closed file!\n");
@@ -1608,33 +1650,14 @@ static void do_debug (int fd)
 	}
 	if (info.overrun_count > 0)
 	    SYSLOG (LOG_ERR, "%u events have been lost!\n",info.overrun_count);
-	switch (info.type)
+	for (count = 0; event_types[count].config_name != NULL; ++count)
 	{
-	  case DEVFSD_NOTIFY_REGISTERED:
-	    type = "registered";
+	    if (info.type != event_types[count].type) continue;
+	    type = event_types[count].debug_name;
 	    break;
-	  case DEVFSD_NOTIFY_UNREGISTERED:
-	    type = "unregistered";
-	    break;
-	  case DEVFSD_NOTIFY_ASYNC_OPEN:
-	    type = "asynchronously opened";
-	    break;
-	  case DEVFSD_NOTIFY_CLOSE:
-	    type = "closed";
-	    break;
-	  case DEVFSD_NOTIFY_LOOKUP:
-	    type = "lookup";
-	    break;
-	  case DEVFSD_NOTIFY_CHANGE:
-	    type = "change";
-	    break;
-	  case DEVFSD_NOTIFY_CREATE:
-	    type = "create";
-	    break;
-	  case DEVFSD_NOTIFY_DELETE:
-	    type = "delete";
-	    break;
-	  default:
+	}
+	if (type == NULL)
+	{
 	    fprintf (stderr, "Unknown type: %u\n", info.type);
 	    exit (1);
 	    /*break;*/
@@ -1715,10 +1738,10 @@ static mode_t get_mode (const char *string)
     return (mode);
 }   /*  End Function get_mode  */
 
-static void sig_handler (int sig)
+static void sighup_handler (int sig)
 {
     SYSLOG (LOG_INFO, "Caught SIGHUP\n");
-}   /*  End Function sig_handler  */
+}   /*  End Function sighup_handler  */
 
 static CONST char *get_variable (CONST char *variable, void *info)
 {
@@ -1804,11 +1827,6 @@ static void do_scan_and_service (CONST char *dirname)
 		    path, ERRSTRING);
 	    continue;
 	}
-	if ( S_ISDIR (statbuf.st_mode) )
-	{
-	    do_scan_and_service (path);
-	    continue;
-	}
 	memset (&info, 0, sizeof info);
 	info.type = DEVFSD_NOTIFY_REGISTERED;
 	info.mode = statbuf.st_mode;
@@ -1820,20 +1838,21 @@ static void do_scan_and_service (CONST char *dirname)
 		  path + strlen (mount_point) + 1);
 	info.namelen = strlen (info.devname);
 	service_name (&info);
+	if ( S_ISDIR (statbuf.st_mode) ) do_scan_and_service (path);
     }
     closedir (dp);
 }   /*  End Function do_scan_and_service  */
 
-static flag make_symlink (const char *oldpath, const char *newpath)
+int mksymlink (const char *oldpath, const char *newpath)
 /*  [SUMMARY] Create a symlink, creating intervening directories as required.
     <oldpath> The string contained in the symlink.
     <newpath> The name of the new symlink.
-    [RETURNS] TRUE on success, else FALSE.
+    [RETURNS] 0 on success, else -1.
 */
 {
-    static char function_name[] = "make_symlink";
+    static char function_name[] = "mksymlink";
 
-    if ( !make_dir_tree (newpath) ) return (FALSE);
+    if ( !make_dir_tree (newpath) ) return (-1);
     if (symlink (oldpath, newpath) != 0)
     {
 	if (errno == EEXIST)
@@ -1842,12 +1861,14 @@ static flag make_symlink (const char *oldpath, const char *newpath)
 		fprintf (stderr, "symlink: \"%s\" already exists\n", newpath);
 	}
 	else
+	{
 	    SYSLOG (LOG_ERR, "%s: error creating symlink: \"%s\"\t%s\n",
 		    function_name, newpath, ERRSTRING);
-	return (FALSE);
+	    return (-1);
+	}
     }
-    return (TRUE);
-}   /*  End Function make_symlink  */
+    return (0);
+}   /*  End Function mksymlink  */
 
 static flag make_dir_tree (const char *path)
 /*  [SUMMARY] Creating intervening directories for a path as required.
